
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    xor(true, true) -> false
%    xor(true, false) ->  true
%    xor(false, true) ->  true
%    xor(false, false) -> false
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%
% Evaluation to normal form (nf).
nf(T, T) :- var(T), !.
nf(T, NfT) :- hnf(T, HnfT), nf1(HnfT, NfT).

nf1(true,true).
nf1(false,false).

%%%%%%%%%
% Evaluation to head normal form (hnf).
hnf(T,T) :- var(T), !.
hnf(xor(X1,X2), H) :- !, xor(X1,X2,H).
hnf(ixor(X1,X2), H) :- !, ixor(X1,X2,H).
hnf(T,T). % the term T is in hnf; a constant is in hnf

%%%%%%%%%
% Clauses to implement the "xor" defined fuction
xor(A,B,C):-hnf(A,F),xor_1(F,B,C).

xor_1(true,A,B):-hnf(A,E),xor_1_true_1(E,B).
xor_1_true_1(true,false).
xor_1_true_1(false,true).

xor_1(false,B,C):-hnf(B,F),xor_1_false_2(F,C).
xor_1_false_2(true,true).
xor_1_false_2(false,false).


%%%%%%%%%%%%%%%%%%%%%%
%%% block IMPROVED xor
ixor(A,B,C):-hnf(A,F), hnf(B,G), ixor_1(F,G,C).

ixor_1(true,true,false).
ixor_1(true,false,true).
ixor_1(false,true,true).
ixor_1(false,false,false).


%%%/**************** SICSTUS
%%%% This is how time and space are measured.
stat_start :- garbage_collect, nogc, statistics(runtime,_).
stat_end(Time) :-
        statistics(runtime,[_,MSec]), Time is MSec/1000,
        statistics(global_stack,[Bytes,_]), Mb is Bytes/1000000,
        width(WT,Time), format("Time ~*c~2f sec   ", [WT, 0' , Time]),
        width(WM,Mb), format("Garb ~*c~3f Mb   ", [WM, 0' , Mb]),
        nl.

% statistics
width(0,X) :- X>=10, !.
width(1,_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

eval(Computation):- nf(Computation,_), fail.
eval(_).


%% g(Input,Input)
%% V is the numer times we want to run the Computation,
g(0,_).
g(V,Computation):- eval(Computation), NewV is V-1, g(NewV,Computation).

%% g0(Input,Output)
%% V is the numer times we want to run the computation,
%% T is the runtime for this computation.
g0(V,T):- write('goal 0: '), stat_start, g(V,_), stat_end(T).

%% g1(Input,Output)
%% Idem
g1(V,T):- write('goal 1: xor  ..'), stat_start,
	g(V,xor(_,_)), stat_end(T).

%% g2(Input,Output)
%% Idem
g2(V,T):- write('goal 2: ixor ..'), stat_start,
	g(V,ixor(_,_)), stat_end(T).

allgoals:- V = 100000,g0(V,T0),g1(V,T1),g2(V,T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl.
%%%*******************************************************/

/**************** SWI-ROLOG
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% T1 is the start time
stat_start(T1) :- T1 is cputime.

%% stat_end(Input,Output)
%% T1 is the starting time and T is the temporal cost of the whole computation,
stat_end(T1,T) :-
	T2 is cputime, T is T2-T1, write('T= '), write(T),
        nl.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



eval(Computation):- nf(Computation,_), fail.
eval(_).


%% g(Input,Input)
%% V is the numer times we want to run the Computation,
g(0,_).
g(V,Computation):- eval(Computation), NewV is V-1, g(NewV,Computation).

%% g0(Input,Output)
%% V is the numer times we want to run the computation,
%% T is the runtime for this computation.
g0(V,T):- write('goal 0: '), stat_start(T1), g(V,_), stat_end(T1,T).

%% g1(Input,Output)
%% Idem
g1(V,T):- write('goal 1: xor  ..'), stat_start(T1),
	g(V,xor(_,_)), stat_end(T1,T).

%% g2(Input,Output)
%% Idem
g2(V,T):- write('goal 2: ixor ..'), stat_start(T1),
	g(V,ixor(_,_)), stat_end(T1,T).

allgoals:- V = 100000,g0(V,T0),g1(V,T1),g2(V,T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl.

**********************************************************/
