%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%  R_1: even(0) ---> 'True', 
%%  R_2: even(succ(succ(X)) --->  even(X).
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Auxiliary predicate to built natural numbers.
built_nat(0, zero) :- !.
built_nat(I, succ(N)) :- NextI is (I - 1), built_nat(NextI, N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block even : CURRY2PROLOG
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Clause for the root node: it exploits the first inductive position
even(X1, H) :- hnf(X1, HX1), even_1(HX1, H).

% Clauses for the remainder nodes:
even_1(zero, 'True').
even_1(succ(X2), H) :- hnf(X2, HX2), even_1_s_1_1(HX2, H).

even_1_s_1_1(succ(X3), H) :- even(X3, H).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block IMPROVED even (exploiting the deterministic branch)
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Clause for the root node: it exploits the first inductive position
ieven(X1, H) :- hnf(X1, HX1), ieven_1(HX1, H).

% Clauses for the remainder nodes:
ieven_1(zero, 'True').
ieven_1(succ(X2), H) :- hnf(X2, succ(X3)),  ieven(X3, H).



%%%%%%%%%%%%%
%%% block hnf
hnf(A,B):-var(A),!,B=A.
hnf(even(A),D):-!,even(A,D).
hnf(ieven(A),D):-!,ieven(A,D).
hnf(A,A).

%%%%%%%%%%%%
%%% block nf
nf(A,B):-hnf(A,E),nf_hnf(E,B).

%%% block nf_hnf
nf_hnf(A,B):-var(A),!,B=A.
nf_hnf('True','True'):-!.
nf_hnf('False','False'):-!.
nf_hnf(zero,zero):-!.
nf_hnf(succ(A),succ(B)):-!,nf(A,B).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Example of goals considered in the experiments %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%% SICSTUS
%%%% This is how time and space are measured.
stat_start :- garbage_collect, nogc, statistics(runtime,_).
stat_end(Time) :-
        statistics(runtime,[_,MSec]), Time is MSec/1000,
        statistics(global_stack,[Bytes,_]), Mb is Bytes/1000000,
        width(WT,Time), format("Time ~*c~2f sec   ", [WT, 0' , Time]),
        width(WM,Mb), format("Garb ~*c~3f Mb   ", [WM, 0' , Mb]),
        nl.

% statistics
width(0,X) :- X>=10, !.
width(1,_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%
g0(T) :- write('goal 0: ...'), stat_start,
      nf(_, _),
      stat_end(T),
      nl.


%%%%%
g1(T) :- write('goal 1:  even(160000) ... '),
      built_nat(160000, N1),
      stat_start,
      nf(even(N1), _),
      stat_end(T),
      nl.

%%%%%
g2(T) :- write('goal 2: ieven(160000) ... '),
      built_nat(160000, N1),
      stat_start,
      nf(ieven(N1), _),
      stat_end(T),
      nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
allgoals:- g0(T0),g1(T1),g2(T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl, nl.





